"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const Endpoint_1 = require("./Endpoint");
const utils_1 = require("./utils");
const serializeError = require("serialize-error");
var RuntimeContext;
(function (RuntimeContext) {
    RuntimeContext[RuntimeContext["Window"] = 0] = "Window";
    RuntimeContext[RuntimeContext["Frame"] = 1] = "Frame";
    RuntimeContext[RuntimeContext["Devtools"] = 2] = "Devtools";
    RuntimeContext[RuntimeContext["Background"] = 3] = "Background";
    RuntimeContext[RuntimeContext["ContentScript"] = 4] = "ContentScript";
    RuntimeContext[RuntimeContext["Worker"] = 5] = "Worker";
})(RuntimeContext || (RuntimeContext = {}));
const ENDPOINT_RE = /^((?:background$)|devtools|content-script|window)(?:@(\d+))?$/;
/**
 * Bridge
 * @external
 */
class Bridge {
    /**
     * Sends a message to some other endpoint, to which only one listener can send response.
     * Returns Promise. Use `then` or `await` to wait for the response.
     * If destination is `window` message will routed using window.postMessage.
     * Which requires a shared namespace to be set between `content-script` and `window`
     * that way they can recognize each other when global window.postMessage happens and there are other
     * extensions using crx-bridge as well
     * @param messageID
     * @param data
     * @param destination
     */
    static sendMessage(messageID, data, destination) {
        let endpoint = (destination instanceof Endpoint_1.Endpoint) ? destination.name : destination;
        const errFn = 'Bridge#sendMessage ->';
        if (!ENDPOINT_RE.test(endpoint)) {
            throw new TypeError(`${errFn} Destination must be any one of known destinations`);
        }
        if (this.context === RuntimeContext.Background) {
            const [input, dest, destTabId] = endpoint.match(ENDPOINT_RE);
            if (dest !== 'background' && !destTabId || !parseInt(destTabId, 10)) {
                throw new TypeError(`${errFn} When sending messages from background page, use @tabId syntax to target specific tab`);
            }
        }
        return new Promise((resolve, reject) => {
            const payload = {
                messageID,
                data,
                destination: endpoint,
                messageType: 'message',
                transactionId: utils_1.uuid(),
                origin: this.ctxname,
                hops: [],
                timestamp: Date.now(),
            };
            this.openTransactions.set(payload.transactionId, { resolve, reject });
            this.routeMessage(payload);
        });
    }
    static onMessage(messageID, callback) {
        this.onMessageListeners.set(messageID, callback);
    }
    static init() {
        if (this.isInitialized) {
            return;
        }
        if (this.context === null) {
            throw new Error(`Unable to detect runtime context i.e crx-bridge can't figure out what to do`);
        }
        this.id = utils_1.uuid();
        this.ctxname = RuntimeContext[this.context].toLowerCase();
        if (this.context === RuntimeContext.Window || this.context === RuntimeContext.ContentScript) {
            window.addEventListener('message', this.handleWindowOnMessage);
        }
        if (this.context === RuntimeContext.ContentScript) {
            this.ctxname = 'content-script';
        }
        if (this.context === RuntimeContext.Devtools) {
            const tabId = chrome.devtools.inspectedWindow.tabId;
            const name = `devtools@${tabId}`;
            this.ctxname = name;
            this.port = chrome.runtime.connect({ name });
            this.port.onMessage.addListener((message) => {
                this.routeMessage(message);
            });
            this.port.onDisconnect.addListener(() => {
                this.port = null;
            });
        }
        if (this.context === RuntimeContext.Background) {
            chrome.runtime.onConnect.addListener((port) => {
                const id = port.name !== '' ? port.name : `content-script@${port.sender.tab.id}`;
                this.portMap.set(id, port);
                this.messageQueue.forEach((queuedMsg) => {
                    if (queuedMsg.resolvedNextNode === id) {
                        port.postMessage(queuedMsg.message);
                        this.messageQueue.delete(queuedMsg);
                    }
                });
                port.onDisconnect.addListener(() => {
                    this.portMap.delete(id);
                });
                port.onMessage.addListener((message) => {
                    this.routeMessage(message, { sourcePort: port });
                });
            });
        }
        if (this.context === RuntimeContext.ContentScript && top === window) {
            this.port = chrome.runtime.connect();
            this.port.onMessage.addListener((message) => {
                this.routeMessage(message);
            });
        }
        this.isInitialized = true;
    }
    static handleInboundMessage(message) {
        return __awaiter(this, void 0, void 0, function* () {
            const { transactionId, messageID, messageType } = message;
            // Sender context
            if (messageType === 'reply') {
                const transactionP = Bridge.openTransactions.get(transactionId);
                if (transactionP) {
                    const { err, data } = message;
                    if (err) {
                        const hydratedErr = new (typeof self[err.name] === 'function' ? self[err.name] : Error)(err.message);
                        for (let prop in err) {
                            hydratedErr[prop] = err[prop];
                        }
                        transactionP.reject(hydratedErr);
                    }
                    else {
                        transactionP.resolve(data);
                    }
                    Bridge.openTransactions.delete(transactionId);
                }
            }
            else if (messageType === 'message') {
                let reply;
                let err;
                let noHandlerFoundError = false;
                try {
                    const cb = Bridge.onMessageListeners.get(messageID);
                    if (typeof cb === 'function') {
                        reply = yield cb({
                            sender: new Endpoint_1.Endpoint(message.origin),
                            id: messageID,
                            data: message.data,
                            timestamp: message.timestamp,
                        });
                    }
                    else {
                        noHandlerFoundError = true;
                        throw new Error(`[crx-bridge] No handler registered in '${Bridge.ctxname}' to accept messages with id '${messageID}'`);
                    }
                }
                catch (error) {
                    err = error;
                }
                finally {
                    if (err) {
                        message.err = serializeError(err);
                    }
                    Bridge.routeMessage(Object.assign({}, message, { messageType: 'reply', data: reply, origin: Bridge.ctxname, destination: message.origin, hops: [] }));
                    if (err && !noHandlerFoundError) {
                        throw reply;
                    }
                }
            }
        });
    }
    static winRouteMsg(win, msg) {
        Bridge.ensureNamespace();
        const channel = new MessageChannel();
        const retry = setTimeout(() => {
            channel.port1.onmessage = null;
            Bridge.winRouteMsg(win, msg);
        }, 300);
        channel.port1.onmessage = (ev) => {
            clearTimeout(retry);
            win.postMessage({
                cmd: '__crx_bridge_route_message',
                scope: Bridge.namespace,
                payload: msg,
                context: Bridge.context,
            }, '*');
        };
        win.postMessage({ cmd: '__crx_bridge_verify_listening', scope: Bridge.namespace, context: Bridge.context }, '*', [channel.port2]);
    }
}
Bridge.ctxname = null;
Bridge.id = null;
Bridge.context = (chrome.devtools) ? RuntimeContext.Devtools
    : (chrome.tabs) ? RuntimeContext.Background
        : (chrome.extension) ? RuntimeContext.ContentScript
            : (typeof document === 'undefined' && typeof importScripts === 'function') ? RuntimeContext.Worker
                : (typeof document !== 'undefined' && window.top !== window) ? RuntimeContext.Frame
                    : (typeof document !== 'undefined') ? RuntimeContext.Window : null;
Bridge.isExternalMessagingEnabled = false;
Bridge.linkedNodesCache = {};
Bridge.isInitialized = false;
Bridge.openTransactions = new Map();
Bridge.onMessageListeners = new Map();
Bridge.port = null;
Bridge.portMap = new Map();
Bridge.messageQueue = new Set();
Bridge.setNamespace = (nsps) => {
    Bridge.namespace = nsps;
};
Bridge.allowWindowMessaging = (nsps) => {
    Bridge.isWindowMessagingAllowed = true;
    Bridge.namespace = nsps;
};
Bridge.handleWindowOnMessage = (ev) => __awaiter(this, void 0, void 0, function* () {
    if (Bridge.context === RuntimeContext.ContentScript && !Bridge.isWindowMessagingAllowed) {
        return;
    }
    const { data, source, ports } = ev;
    if (data.cmd === '__crx_bridge_verify_listening' && data.scope === Bridge.namespace && data.context !== Bridge.context) {
        const port = ports[0];
        port.postMessage(true);
        return;
    }
    else if (data.cmd === '__crx_bridge_route_message' && data.scope === Bridge.namespace && data.context !== Bridge.context) {
        // a message event insdide `content-script` means a script inside `window` dispactched it
        // so we're making sure that the origin is not tampered (i.e script is not masquerading it's true identity)
        if (Bridge.context === RuntimeContext.ContentScript) {
            data.payload.origin = 'window';
        }
        Bridge.routeMessage(data.payload);
    }
});
Bridge.routeMessage = (message, options = {}) => __awaiter(this, void 0, void 0, function* () {
    const { origin, messageID, transactionId, destination } = message;
    if (message.hops.indexOf(Bridge.id) > -1) {
        return;
    }
    message.hops.push(Bridge.id);
    if (Bridge.context === RuntimeContext.ContentScript
        && /window/.test(message.destination + origin)
        && !Bridge.allowWindowMessaging) {
        return;
    }
    // It's previous node's job to unset destination, before handing over the payload
    if (!destination) {
        Bridge.handleInboundMessage(message);
    }
    if (ENDPOINT_RE.test(destination)) {
        if (Bridge.context === RuntimeContext.Frame) {
            throw new Error(`crx-bridge no longer supports iframes due to too much complexity and not so many use cases`);
        }
        else if (Bridge.context === RuntimeContext.Window) {
            Bridge.winRouteMsg(window, message);
        }
        else if (Bridge.context === RuntimeContext.ContentScript && destination === 'window') {
            message.destination = null;
            Bridge.winRouteMsg(window, message);
        }
        else if (Bridge.context === RuntimeContext.Devtools || Bridge.context === RuntimeContext.ContentScript) {
            if (/background/.test(destination)) {
                message.destination = null;
            }
            // Just hand it over to background page
            Bridge.port.postMessage(message);
        }
        else if (Bridge.context === RuntimeContext.Background && options.sourcePort) {
            const [dest, destName, destTabId] = destination.match(ENDPOINT_RE);
            const [src, srcName, srcTabId = `${options.sourcePort.sender.tab.id}`] = origin.match(ENDPOINT_RE);
            if (/content-script|window/.test(srcName)) {
                message.origin = `${srcName}@${srcTabId}`;
            }
            const fixedDest = destTabId ? dest : `${dest}@${srcTabId}`;
            if (destName !== 'window') {
                message.destination = null;
            }
            else {
                // we're in background page, hence at this point the tab id has been resolved already, so in case it's like `window@932`
                // we set this to just `window` because that let's content script know that this is the end of payload journey and stops
                // it from recursive message transfers
                message.destination = 'window';
            }
            // As far as background page is concerned, it just needs to get the payload to "actual" tab
            const resolvedNextNode = destName === 'window' ? fixedDest.replace('window', 'content-script') : fixedDest;
            const port = Bridge.portMap.get(resolvedNextNode);
            if (port) {
                port.postMessage(message);
            }
            else {
                Bridge.messageQueue.add({ resolvedNextNode, message });
            }
        }
    }
});
Bridge.ensureNamespace = () => {
    if (typeof Bridge.namespace !== 'string' || Bridge.namespace.length === 0) {
        const err = `crx-bridge uses window.postMessage to talk with other "window"(s), for message routing and stuff,` +
            ` which is global/conflicting operation in case there are other scripts using crx-bridge. ` +
            `Call Bridge#setNamespace(nsps) to isolate your app. Example: Bridge.setNamespace('com.facebook.react-devtools'). ` +
            `Make sure to use same namespace across all your scripts whereever window.postMessage is likely to be used`;
        throw new Error(err);
    }
};
exports.Bridge = Bridge;
Bridge.init();
//# sourceMappingURL=Bridge.js.map